Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq path div choice.
Require Import fintype tuple finfun bigop prime finset binomial.
Require Import ssralg ssrnum ssrint intdiv rat.

Require Import bigenough.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Import GRing.Theory.
Import Num.Theory.
Import BigEnough.

Open Scope ring_scope.

(* FIXME : are these lemmas missing in MathComp? *)

(* Suggestions for alternative statements of lemmas in big_op: *)
Section AltBigOp.

Variables  (R : Type) (idx : R) (op : Monoid.law idx).

Lemma big_nat_recr_alt (n m : nat) (F : nat -> R) : (m <= n)%N ->
\big[op/idx]_(m <= i < n.+1) F i = op (\big[op/idx]_(m <= i < n) F i) (F n).
Proof. by move=> lemn; rewrite (@big_cat_nat _ _ _ n) //= big_nat1. Qed.

Lemma big_nat_recl_alt (n m : nat) (F : nat -> R) : (m <= n)%N ->
  \big[op/idx]_(m <= i < n.+1) F i =
     op (F m) (\big[op/idx]_(m <= i < n) F i.+1).
Proof. by move=> lemn; rewrite big_ltn // big_add1. Qed.

End AltBigOp.

Section ExtraSsrAlg.

Lemma expfV (R : idomainType) (x : R) (i : int) : (x ^ i) ^-1 = (x ^-1) ^ i.
Proof. by rewrite invr_expz exprz_inv. Qed.

End ExtraSsrAlg.

Section ExtraInt.


(* Two technical lemmas about rationals that are integers *)
(* FIXME: They could be turned into equivalences although this is not needed*)
(* here.*)
Lemma Qnat_div (m d : nat) : (d %| m)%N -> (m%:R / d%:R : rat) \is a Qnat.
Proof.
move/(char0_natf_div (char_num [numDomainType of rat]))<-.
exact: rpred_nat.
Qed.

Lemma Qint_div (m d : int) : (d %| m)%Z -> (m%:~R / d%:~R : rat) \is a Qint.
Proof.
move=> dvddm.
have: `|m%:~R / d%:~R| \is a Qint.
  rewrite normrM normfV -!intr_norm.
  by move: dvddm; rewrite dvdzE /=; move/Qnat_div; rewrite Qnat_def; case/andP.
by case: ler0P (m%:~R / d%:~R : int) => // _ _ h; rewrite -[_ / _]opprK rpredN.
Qed.

End ExtraInt.

Section ExtraRat.

Lemma denqVz (i : int) : i != 0 -> denq (i%:~R^-1) = `|i|.
Proof.
by move=> h; rewrite -div1r -[1]/(1%:~R) coprimeq_den /= ?coprime1n // (negPf h).
Qed.

Lemma ler1z (R : numDomainType) (n : int) : (1 <= n%:~R :> R) = (1 <= n).
Proof. by rewrite -[1]/(1%:~R) ler_int. Qed.

Lemma ltr1z (R : numDomainType) (n : int) : (1 < n%:~R :> R) = (1 < n).
Proof. by rewrite -[1]/(1%:~R) ltr_int. Qed.

End ExtraRat.

Section ExtraSsrNum.

Variable R : numDomainType.

Lemma ler1n (n : nat) : (1 <= n%:~R :> R) = (1 <= n)%N.
Proof. by rewrite -[1]/(1%:~R) ler_nat. Qed.

Lemma ltr1n (n : nat) : (1 < n%:~R :> R) = (1 < n)%N.
Proof. by rewrite -[1]/(1%:~R) ltr_nat. Qed.

Lemma lt0r_neq0 (x : R) : 0 < x  -> x != 0.
Proof. by rewrite lt0r; case/andP. Qed.

Lemma ltr0_neq0 (x : R) : 0 < x  -> x != 0.
Proof. by rewrite lt0r; case/andP. Qed.

Lemma le0r_lt0_add (y t : R) : 0 <= y -> 0 < t -> 0 < y + t.
Proof. move=> le0y lt0t; rewrite -[0]add0r; exact: ler_lt_add. Qed.

Lemma lt0r_le0_add (y t : R) : 0 < y -> 0 <= t -> 0 < y + t.
Proof. move=> lt0y le0t; rewrite -[0]add0r; exact: ltr_le_add. Qed.

Lemma expN1r (i : int) : (-1 : R) ^ i = (-1) ^+ `|i|.
Proof.
case: i => n; first by rewrite exprnP absz_nat.
by rewrite NegzE abszN  absz_nat -invr_expz expfV invrN1.
Qed.

(* Can't we shorten this proof ? *)
Lemma ltr_prod (E1 E2 : nat -> R) (n m : nat) :
   (m < n)%N -> (forall i, (m <= i < n)%N -> 0 <= E1 i < E2 i) ->
  \prod_(m <= i < n) E1 i < \prod_(m <= i < n) E2 i.
Proof.
wlog: E1 E2 n m / m = 0%N.
  move=> hwlog ltmn hE12.
  have -> : (m = 0 + m)%N by rewrite add0n.
  rewrite !big_addn; apply: hwlog => //; first by rewrite subn_gt0.
  by move=> i hi; apply: hE12; rewrite leq_addl addnC -ltn_subRL.
move=> -> {m}; elim: n => //; case=> [_ |n ihn] _ hE12.
- by rewrite !big_nat_recl !big_nil !mulr1; case/andP: (hE12 0%N (erefl true)).
rewrite [in X in X < _]big_nat_recr [in X in _ < X]big_nat_recr /=.
have /hE12 /andP[le0E1 lt12]: (0 <= n.+1 < n.+2)%N by rewrite ltnS leq0n leqnn.
apply: ltr_pmul => //.
- rewrite big_nat_cond; apply: prodr_ge0 => i; rewrite andbT => /andP[_ lein].
  suff : (0 <= i < n.+2)%N by move/hE12 => /andP [].
  by rewrite leq0n /= ltnW.
- by apply: ihn => // i /andP[_ ltin]; apply: hE12; rewrite leq0n ltnW.
Qed.

End ExtraSsrNum.

(*** Two lemmas about geometric sequences. We just use the second one. ***)
Section ExtraGeom.

(* FIXME : these two lemmas are not cleaned up *)
(* The proof uses bigenough so that's not strictly in MathComp, but *)
(* as usual we could eliminate the pose_big_enough a posteriori. *)
Lemma Gseqgt1 (r : rat)(E : rat) : 0 <= E -> 1 < r ->
  exists N : nat, forall n : nat, (N <= n)%nat -> E < r ^ n.
Proof.
move=> ge0E lt1r.
pose_big_enough M.
  exists M => k hMk.
  have le1r : 1 <= r by rewrite ltrW.
  have {lt1r} lt1r : 0 < r - 1 by rewrite subr_gt0.
  suff aux (n : nat) : n%:~R * (r - 1) - 1 <= r ^ n.
    apply: ltr_le_trans (aux k); rewrite -ltr_sub_addr opprK -ltr_pdivr_mulr //.
    have h : 0 <= (E + 1) / (r - 1).
       apply: divr_ge0; by [exact:  addr_ge0 | exact: ltrW].
    by apply: ltr_trans (archi_boundP h) _; rewrite ltr_nat.
  elim: n => [|n ihn]; first by rewrite mul0r add0r expr0z.
  have -> : n.+1%:~R * (r - 1) - 1 = n%:~R * (r - 1) - 1 + (r - 1).
    by rewrite -addn1 PoszD rmorphD /= mulrDl mul1r addrAC.
  have -> : r ^ n.+1 = r ^ n + (r ^ n.+1 - r ^ n) by rewrite addrCA addrN addr0.
  apply: ler_add => //.
  have -> : r ^ n.+1 - r ^ n = r ^ n * (r - 1).
    by rewrite mulrDr mulrN mulr1 -exprnP exprSr.
  case: n {ihn} => [| n ]; first by rewrite mul1r.
  rewrite ler_pmull // -exprnP expr_ge1 //; exact: ler_trans le1r.
by close.
Qed.

Lemma Gseqlt1 (r : rat)(eps : rat) : 0 < eps -> 0 < r < 1 ->
  exists N : nat, forall n : nat, (N <= n)%nat -> r ^ n < eps.
Proof.
move=> ge0eps /andP [lt0r ltr1].
have hE : 0 <= eps ^-1 by apply/ltrW; rewrite invr_gt0.
have hr : 1 < r ^-1 by rewrite invf_gt1.
have [M hM] := Gseqgt1 hE hr; exists M => n hMn.
rewrite -ltf_pinv //; first by rewrite invr_expz -exprz_inv; exact: hM.
exact: exprz_gt0.
Qed.

End ExtraGeom.


Section Telescope.

Variable R : ringType.

Lemma telescope_nat (a b : nat) (f : nat -> R) : (a <= b)%N ->
  \sum_(a <= k < b) (f (k + 1)%N - f k) = f b - f a.
Proof.
rewrite -{2}[a]add0n big_addn; elim: b => [ | b ihb].
  by rewrite leqn0=> /eqP ->; rewrite subrr subnn big_mkord big_ord0.
rewrite leq_eqVlt; case/orP=> [/eqP -> | a_lt_b1].
  by rewrite subnn subrr big_geq.
rewrite subSn ?big_nat_recr //= ihb ?subnK // addn1 addrC -addrA [- _ + _]addrA.
by rewrite addNr sub0r.
Qed.

End Telescope.


Section TelescopeProd.

Variable R : fieldType.

Lemma telescope_prod_nat (a b : nat) (f : nat -> R) :
 (forall k, (a <= k < b)%N -> f k != 0) ->  (a < b)%N ->
  \prod_(a <= k < b) (f (k + 1)%N / f k) = f b / f a.
Proof.
move=> hf; rewrite -{2}[a]add0n big_addn; elim: b hf => [ | b ihb] hf //.
rewrite ltnS leq_eqVlt; case/orP=> [/eqP -> | a_lt_b1].
  by rewrite subSn // subnn big_nat_recr /= big_nil mul1r add0n addn1.
rewrite subSn ?big_nat_recr //= ihb ?subnK //; last first.
  by move=> k /andP [hak hkb]; apply: hf; rewrite hak ltnS ltnW. 
rewrite addn1 -!mulrA mulrC -!mulrA mulVf; first by rewrite mulr1 mulrC.
by apply: hf; rewrite ltnW //=.
Qed.

End TelescopeProd.
